@@ -21,6 +21,17 @@ aembit_nat_OUTPUT=aembit_nat_OUTPUT
 aembit_nat_PREROUTING=aembit_nat_PREROUTING
 aembit_filter_INPUT=aembit_filter_INPUT
 
+# Enable dynamic routing if the hosts were specified.
+dynamic_routing=false
+if [ -n "${AEMBIT_STEERING_HOSTS}" ]; then
+    dynamic_routing=true
+fi
+
+log() {
+    timestamp=$(date '+%Y-%m-%d %H:%M:%S')
+    echo "[${timestamp}] $1"
+}
+
 # Create all aembit rule chains.
 create_all_aembit_chains() {
     add_ipv4_chain raw "${aembit_raw_PREROUTING}"
@@ -240,12 +251,70 @@ function redirect_outbound_tcp_traffic_to_ap() {
     # See https://learn.microsoft.com/en-us/azure/virtual-network/what-is-ip-address-168-63-129-16.
     add_ipv4_rule nat "${aembit_nat_OUTPUT}" -p tcp --destination "${azure_platform_resources_ip}" -j RETURN
 
-    # Redirect all new TCP connections to the Agent Proxy.
-    # We use the "--syn" flag to forward new connections only, since,
-    # prior to adding NAT rules, conntrack is not loaded. Once we add NAT rules,
-    # conntrack is loaded, but mistakenly identifies existing connections as new
-    # connections, redirecting them erroneously to the Agent Proxy mid-connection.
-    add_ipv4_rule nat "${aembit_nat_OUTPUT}" -p tcp --syn -j REDIRECT --to-port "${agent_proxy_port}"
+    if [ "${dynamic_routing}" = false ]; then
+        # Redirect all new TCP connections to the Agent Proxy if we're not routing dynamically.
+        #
+        # We use the "--syn" flag to forward new connections only, since,
+        # prior to adding NAT rules, conntrack is not loaded. Once we add NAT rules,
+        # conntrack is loaded, but mistakenly identifies existing connections as new
+        # connections, redirecting them erroneously to the Agent Proxy mid-connection.
+        add_ipv4_rule nat "${aembit_nat_OUTPUT}" -p tcp --syn -m comment --comment global -j REDIRECT --to-port "${agent_proxy_port}"
+    fi
+}
+
+# Manipulates the outbound traffic rules to redirect traffic to the Agent Proxy.
+#
+# This function expects arguments to be:
+# $1: hostname,
+# $@: list of IPs which the hostname resolved to.
+function redirect_outbound_tcp_traffic_for_ips_to_ap() {
+    hostname="$1"
+    shift
+    incoming_ips_for_host=("$@")
+    active_rules=$(iptables -t nat -S ${aembit_nat_OUTPUT} | grep -e "-m comment --comment \"${hostname}\"")
+    active_rules=${active_rules//\"}
+    IFS=$'\n' read -rd '' -a active_rules_for_host <<< "${active_rules}"
+
+    # We need to do three things:
+    # 1. Remove any existing rules that are stale (i.e. DNS resolution for this hostname doesn't include the ip anymore),
+    # 2. Add any new rules for IPs that are not already in the active rules,
+    # 3. Don't remove rules that are still valid (i.e. there's a rule for the incoming IP in the active rules).
+    #
+    # We can do this by taking the intersection of the incoming IPs and the active rules, and then removing the IPs/rules that are in both.
+    # Since bash doesn't have sets, we'll just loop over the incoming IPs and remove each entry manually. It's n^2, but the lists are short =).
+    #
+    # At the end, whatever's left in active_rules_for_host gets removed, and whatever's in incoming_ips_for_hosts gets added.
+
+    # For each index of the incoming_ips...
+    for ip_index in "${!incoming_ips_for_host[@]}"; do
+        ip="${incoming_ips_for_host[$ip_index]}"
+
+        # ...and for each index of the active_rules...
+        for rule_index in "${!active_rules_for_host[@]}"; do
+            rule="${active_rules_for_host[$rule_index]}"
+
+            # ...if the rule contains the fragment `-d ${ip}/32`, unset the rule and the IP to remove them from their respective arrays
+            if [[ "$rule" == *"-d ${ip}/32"* ]]; then
+                unset 'active_rules_for_host[rule_index]'
+                unset 'incoming_ips_for_host[ip_index]'
+                break
+            fi
+        done
+    done
+
+    # Finally, remove and add rules
+    for rule in "${active_rules_for_host[@]}"; do
+        # removal expects "<table> <chain> [...]"", but the current format is "-A <chain> [...]"
+        rule="${rule//-A /nat }"
+        log "removing: ${rule}"
+        IFS=' ' read -ra rule_parts <<< "${rule}"
+        remove_ipv4_rule "${rule_parts[@]}"
+    done
+    for ip in "${incoming_ips_for_host[@]}"; do
+        rule_parts=("nat" "${aembit_nat_OUTPUT}" "-d" "${ip}" "-p" "tcp" "--syn" "-m" "comment" "--comment" "${hostname}" "-j" "REDIRECT" "--to-port" "${agent_proxy_port}")
+        log "adding: ${rule_parts[*]}"
+        add_ipv4_rule "${rule_parts[@]}"
+    done
 }
 
 # Redirect TCP traffic from containers to the AP.
@@ -288,6 +357,7 @@ function install_traffic_steering_rules_to_ap() {
     save_and_reload_daemon
 }
 
+
 # Uninstall chain rules for steering traffic to the Agent Proxy.
 # These function wrappers would in turn remove the iptables or firewalld specific rules
 # based upon whether firewalld is active and running.
@@ -299,12 +369,20 @@ function uninstall_traffic_steering_rules_to_ap() {
 case "${OPTION}" in
     install)
         echo "Installing traffic steering rules to the Agent Proxy."
+        shift
         install_traffic_steering_rules_to_ap
         ;;
     uninstall)
         echo "Uninstalling traffic steering rules."
         uninstall_traffic_steering_rules_to_ap
         ;;
+    update)
+        shift
+        # Remaining args are expected to be <hostname> <ip>+, e.g.
+        # ./rules.sh update graph.microsoft.com 10.10.10.10 11.11.11.11 [...]
+        echo "Updating traffic steering rules"
+        redirect_outbound_tcp_traffic_for_ips_to_ap "$@"
+        ;;
     *)
         echo "Invalid option - ${OPTION}"
         ;;
